@@ -35,10 +35,16 @@ unsigned long long parse_num_blocks2(const char *arg, int log_block_size)
 		num <<= 10;
 		/* fallthrough */
 	case 'K': case 'k':
-		num >>= log_block_size;
+		if (log_block_size < 0)
+			num <<= 10;
+		else
+			num >>= log_block_size;
 		break;
 	case 's':
-		num >>= (1+log_block_size);
+		if (log_block_size < 0)
+			num << 1;
+		else
+			num >>= (1+log_block_size);
 		break;
 	case '\0':
 		break;
@@ -62,11 +68,21 @@ main(int argc, char **argv)
 	unsigned long num;
 	int log_block_size = 0;
 
-	if (argc != 2) {
-		fprintf(stderr, "Usage: %s arg\n", argv[0]);
+	if (argc != 2 && argc != 3) {
+		fprintf(stderr, "Usage: %s arg [log_block_size]\n", argv[0]);
 		exit(1);
 	}
 
+	if (argc == 3) {
+		char *p;
+
+		log_block_size = strtol(argv[2], &p, 0);
+		if (*p) {
+			fprintf(stderr, "Bad log_block_size: %s\n", argv[2]);
+			exit(1);
+		}
+	}
+
 	num = parse_num_blocks(argv[1], log_block_size);
 
 	printf("Parsed number: %lu\n", num);
